package java.lang.reflect;

import java.util.HashMap;
import sun.reflect.annotation.AnnotationSupport;
import java.util.Objects;
import java.lang.annotation.Annotation;
import java.util.Map;

public final class Parameter implements AnnotatedElement
{
    private final String name;
    private final int modifiers;
    private final Executable executable;
    private final int index;
    private transient volatile Type parameterTypeCache;
    private transient volatile Class<?> parameterClassCache;
    private transient Map<Class<? extends Annotation>, Annotation> declaredAnnotations;
    
    Parameter(String name, int modifiers, Executable executable, int index) {
        this.parameterTypeCache = null;
        this.parameterClassCache = null;
        this.name = name;
        this.modifiers = modifiers;
        this.executable = executable;
        this.index = index;
    }
    
    @Override
    public boolean equals(Object o) {
        if ((o instanceof Parameter)) {
            Parameter parameter = (Parameter)o;
            return parameter.executable.equals(this.executable) && parameter.index == this.index;
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return this.executable.hashCode() ^ this.index;
    }
    
    public boolean isNamePresent() {
        return this.executable.hasRealParameterData() && this.name != null;
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        String typeName = this.getParameterizedType().getTypeName();
        sb.append(Modifier.toString(this.getModifiers()));
        if (0 != this.modifiers) {
            sb.append(' ');
        }
        if (this.isVarArgs()) {
            sb.append(typeName.replaceFirst("\\[\\]$", "..."));
        }
        else {
            sb.append(typeName);
        }
        sb.append(' ');
        sb.append(this.getName());
        return sb.toString();
    }
    
    public Executable getDeclaringExecutable() {
        return this.executable;
    }
    
    public int getModifiers() {
        return this.modifiers;
    }
    
    public String getName() {
        if (this.name == null || this.name.equals("")) {
            return "arg" + this.index;
        }
        return this.name;
    }
    
    String getRealName() {
        return this.name;
    }
    
    public Type getParameterizedType() {
        Type parameterTypeCache = this.parameterTypeCache;
        if (null == parameterTypeCache) {
            parameterTypeCache = this.executable.getAllGenericParameterTypes()[this.index];
            this.parameterTypeCache = parameterTypeCache;
            return parameterTypeCache;
        }
        return parameterTypeCache;
    }
    
    public Class<?> getType() {
        Class<Object> parameterClassCache = this.parameterClassCache;
        if (null == parameterClassCache) {
            parameterClassCache = this.executable.getParameterTypes()[this.index];
            this.parameterClassCache = parameterClassCache;
            return parameterClassCache;
        }
        return parameterClassCache;
    }
    
    public AnnotatedType getAnnotatedType() {
        return this.executable.getAnnotatedParameterTypes()[this.index];
    }
    
    public boolean isImplicit() {
        return Modifier.isMandated(this.getModifiers());
    }
    
    public boolean isSynthetic() {
        return Modifier.isSynthetic(this.getModifiers());
    }
    
    public boolean isVarArgs() {
        return this.executable.isVarArgs() && this.index == this.executable.getParameterCount() - 1;
    }
    
    @Override
    public <T extends Annotation> T getAnnotation(Class<T> obj) {
        Objects.requireNonNull(obj);
        return (T)obj.cast(this.declaredAnnotations().get(obj));
    }
    
    @Override
    public <T extends Annotation> T[] getAnnotationsByType(Class<T> clazz) {
        Objects.requireNonNull(clazz);
        return AnnotationSupport.getDirectlyAndIndirectlyPresent(this.declaredAnnotations(), clazz);
    }
    
    @Override
    public Annotation[] getDeclaredAnnotations() {
        return this.executable.getParameterAnnotations()[this.index];
    }
    
    @Override
    public <T extends Annotation> T getDeclaredAnnotation(Class<T> clazz) {
        return this.getAnnotation(clazz);
    }
    
    @Override
    public <T extends Annotation> T[] getDeclaredAnnotationsByType(Class<T> clazz) {
        return this.getAnnotationsByType(clazz);
    }
    
    @Override
    public Annotation[] getAnnotations() {
        return this.getDeclaredAnnotations();
    }
    
    private synchronized Map<Class<? extends Annotation>, Annotation> declaredAnnotations() {
        if (null == this.declaredAnnotations) {
            this.declaredAnnotations = new HashMap();
            Annotation[] declaredAnnotations = this.getDeclaredAnnotations();
            for (int i = 0; i < declaredAnnotations.length; ++i) {
                this.declaredAnnotations.put(declaredAnnotations[i].annotationType(), declaredAnnotations[i]);
            }
        }
        return this.declaredAnnotations;
    }
}
